/* See {iagraph.h} */
/* Last edited on 2007-02-15 01:00:38 by stolfi */

#include <iagraph.h>

#include <flt.h>
#include <ia.h>
#include <ia_butfly.h>
#include <ps.h>
#include <affirm.h>

#include <math.h>
#include <stdio.h>

void iagraph_plot_boxes
  ( FILE *psfile,
    Interval f (Interval x),
    Interval xd,
    Interval yd,
    int n
  )
  { Interval xv, yv;
    int xi;
    double gray = 0.75;

    ps_comment(psfile, "Function plot with IA box enclosures");

    for (xi=0; xi<n; xi++)
      {
        ROUND_DOWN;
        xv.lo = xd.lo + ((xd.hi - xd.lo)*xi)/n;

        ROUND_UP;
        xv.hi = xd.lo + ((xd.hi - xd.lo)*(xi+1))/n;

        yv = f(xv);

        ROUND_NEAR;
        
        iagraph_fill_and_draw_box(psfile, xv, yv, xd, yd, gray,gray,gray);
      }

    fprintf(stderr, "\n");
  }

void iagraph_plot_butterflies
  ( FILE *psfile,
    Interval f (Interval x),
    Interval df (Interval x),
    Interval xd,
    Interval yd,
    int n
  )
  { Interval xv, yvlo, yvmd, yvhi;
    int xi;
    double gray = 0.75;

    ps_comment(psfile, "Function plot IA interval-slope enclosures");

    for (xi=0; xi<n; xi++)
      {
        ROUND_DOWN;
        xv.lo = xd.lo + ((xd.hi - xd.lo)*xi)/n;

        ROUND_UP;
        xv.hi = xd.lo + ((xd.hi - xd.lo)*(xi+1))/n;
        
        ROUND_NEAR;

        Float xmd = ia_mid(xv);
        iagraph_compute_butterfly(xv, xmd, f, df, &bt);
        
        iagraph_fill_and_draw_butterfly(psfile, &bt, xd, yd, gray,gray,gray);
      }

    fprintf(stderr, "\n");
  }

void iagraph_compute_butterfly
  ( Interval xv, 
    Float xm, 
    Interval f (Interval x),
    Interval df (Interval x),
    ia_butfly_t *bt
  )
  { demand((xm >= xv.lo) && (xm <= xv.hi), "xm outside xv");
    
    /* Set up X ranges: */
    bt->tp[0].x = (Interval){ xv.lo, xm };
    bt->tp[1].x = (Interval){ xm, xv.hi };
    
    /* Compute slope range over whole range {xv}: */
    Interval dfv = df(xv);

    /* Compute Y range {fxm} at the goven point {xm} of {xv}: */
    Interval xmv = (Interval){xm, xm};
    Interval fxmv = f(xmv);

    #if DEBUG
      fprintf(stderr, "  f(xm) = "); ia_print(stderr, fxmv); fprintf(stderr, "\n");
      fprintf(stderr, "  df(x) = "); ia_print(stderr, dfv); fprintf(stderr, "\n");
    #endif

    if (ia_is_full(&dfv) || ia_is_full(&fxmv))
      { /* Center interval or slope is unbounded; butterfly is unbounded too: */
        bt->tp[0].yxlo = bt->tp[0].yxhi = ia_full();
        bt->tp[1].yxlo = bt->tp[1].yxhi = ia_full();
        return;
      }
    
    /* Compute trapezoids from central value range {yxm} and slope range {dfv}: */
    bt->tp[0].yxhi = bt->tp[1].yxlo = fxmv;
    ROUND_UP;
    double dxlo = xm - xv.lo;
    double dxhi = xv.hi - xm;
    bt->tp[0].yxlo.hi = fxmv.hi + dfv.lo * (-dxlo);
    bt->tp[1].yxhi.hi = fxmv.hi + dfv.hi * (+dxhi);
    ROUND_DOWN;
    bt->tp[0].yxlo.lo = fxmv.lo + dfv.hi * (-dxlo);
    bt->tp[1].yxhi.lo = fxmv.lo + dfv.lo * (+dxhi);
  }

void iagraph_fill_and_draw_box
  ( FILE *psfile, 
    Interval xv, 
    Interval yv, 
    Interval xd,
    Interval yd,
    double R, double G, double B
  )
  { /* Clip box vertically to {yd}: */
    yv = ia_meet(yv, yd);
    /* Plot box: */
    ps_fill_and_draw_rectangle(psfile, xv.lo, xv.hi, yv.lo, yv.hi, R,G,B);
  }

void iagraph_fill_and_draw_trapezoid
  ( FILE *psfile, 
    Interval xv, 
    Interval yvlo, 
    Interval yvhi, 
    Interval xd,
    Interval yd,
    double R, double G, double B
  )
  {
    /* Clip trapezoid vertically against {xd}. May result in hexagon. */
    if (ia_is_full(&yvlo) || ia_is_full(&yvhi)) { yvlo = yvhi = yd; }

    if 
      ( ((yvlo.lo < yd.hi) || (yvhi.lo < yd.hi)) &&
        ((yvlo.hi > yd.lo) || (yvhi.hi > yd.lo))
      ) 
      {
        /* Some part of the trapezoid is visible, plot it: */
        double xp[6], yp[6];
        int np = 0;  /* Number of plot corners. */
        xp[0] = xv.lo;  yp[0] = yvlo.lo;
        xp[1] = xv.hi;  yp[1] = yvhi.lo;
        xp[2] = xv.hi;  yp[2] = yvhi.hi;
        xp[3] = xv.lo;  yp[3] = yvlo.hi;
        np = 4;
        /* Should clip polygon against {xd  yd}: */
        ps_fill_and_draw_polygon(psfile, xp, yp, np, R,G,B); 
      }
  }

void iagraph_fill_and_draw_butterfly
  ( FILE *psfile, 
    ia_butfly_t *bt, 
    Interval xd,
    Interval yd,
    double R, double G, double B
  )
  {
    if (xm > xv.lo) 
      { Interval xa = {xv.lo, xmd};
        iagraph_fill_and_draw_trapezoid(psfile, xa, yvlo, yvmd, xd, yd, R,G,B);
      }
    if (xmd < xv.hi) 
      { Interval xb = {xmd, xv.hi};
        iagraph_fill_and_draw_trapezoid(psfile, xb, yvmd, yvhi, xd, yd, R,G,B);
      }
  }
