/* A short program to check several detaisl of 
   IEEE floating-point arithmetic:
     - whether (-oo) + (+oo) = NaN in all rounding modes.
     - how negative zero behaves with operations.
*/
/* Last edited on 2007-01-04 04:09:52 by stolfi */

#include <ioprotos.h>
#include <stdlib.h>
#include <math.h>
 
// #if (defined(SunOS5))
// 
// #include <ieeefp.h>
// 
// #define ROUND_NEAR { fpsetround(FP_RN); }
// #define ROUND_UP   { fpsetround(FP_RP); }
// #define ROUND_DOWN { fpsetround(FP_RM); }
// 
// #endif

#include <fenv.h>
#include <fpu_control.h>

#define ROUND_NEAR { fesetround(FE_TONEAREST); }
#define ROUND_UP   { fesetround(FE_UPWARD); }
#define ROUND_DOWN { fesetround(FE_DOWNWARD); }

void prt_1_bool_t(char *title, int m, int p);
void prt_1_double(char *title, double m, double p);
void prt_2_bool_t(char *title, int mm, int mp, int pm, int pp);
void prt_2_double(char *title, double mm, double mp, double pm, double pp);

int main()
{
  { /* Testing whether MAXDOUBLE and HUGE_VAL are infinite: */
    fprintf(stderr, "MAXDOUBLE = %24.16e\n", MAXDOUBLE);
    fprintf(stderr, "HUGE_VAL =  %24.16e\n", HUGE_VAL);
    fprintf(stderr, "INF =       %24.16e\n", INF);
  }
  
  { /* Testing whether exp(x) = +oo or finite on overflow: */
    double t, dn, ne, up;
    int i;
    for ( i = 500; i <= 1000; i = i+10 )
      {
        ROUND_NEAR;
        t = i;
        fprintf(stderr, "exp(%f):\n", t);
        ROUND_NEAR;
        ne = exp(t);
        fprintf(stderr, "  NE = %20.12e\n", ne); 
        ROUND_DOWN;
        dn = exp(t);
        fprintf(stderr, "  DN = %20.12e (%20.12e)\n", dn, dn-ne); 
        ROUND_UP;
        up = exp(t);
        fprintf(stderr, "  UP = %20.12e (%20.12e)\n", up, up-ne);
        if ( ne >= up )  fprintf(stderr, "  NE >= UP\n");
        if ( ne <= dn )  fprintf(stderr, "  NE <= DN\n");
      }
  }
  
  { /* Testing whether (-oo) + (+oo) = NaN in all rounding modes: */
    double one, zero, pinf, minf, x, y, z;
    one = 1.0;
    zero = 0.0;
    ROUND_UP;
    pinf = one/zero;
    minf = -pinf;
    ROUND_DOWN;
    x = minf + pinf;
    ROUND_UP;
    y = minf + pinf;
    ROUND_NEAR;
    z = minf + pinf;
    fprintf(stderr, "(-oo)+(+oo): DN = %g  NE = %g UP = %g\n", x, y, z);
  }
  
  { /* Testing the behavior of negative zero in various ops: */
    double mz, pz;
    int i;
    pz = 0.0;
    mz = -pz;
    for (i = 0; i < 3; i++)
      {
        if      (i == 0) { ROUND_NEAR; fprintf(stderr, "\n=== ROUND_NEAR ===\n\n"); }
        else if (i == 1) { ROUND_UP;   fprintf(stderr, "\n=== ROUND_UP ===\n\n"); }
        else if (i == 2) { ROUND_DOWN; fprintf(stderr, "\n=== ROUND_DOWN ===\n\n"); }
        
        prt_1_double("neg",  -mz, -pz);
        prt_1_double("sqrt", sqrt(mz), sqrt(pz));
        prt_1_double("ln",   log(mz), log(pz));
        prt_1_double("inv",  1.0/mz, 1.0/pz);
        prt_2_double("add",  mz+mz, mz+pz, pz+mz, pz+pz);
        prt_2_double("sub",  mz-mz, mz-pz, pz-mz, pz-pz);
        prt_2_double("mul",  mz*mz, mz*pz, pz*mz, pz*pz);

        prt_2_double("eql",  (mz==mz), (mz==pz), (pz==mz), (pz==pz));
        prt_2_double("gtr",  (mz>mz), (mz>pz), (pz>mz), (pz>pz));

      }        
  }
  return 0;
}

void prt_1_bool_t(char *title, int m, int p)
{
  fprintf(stderr, "\n");
  fprintf(stderr, "%s\n", title);
  fprintf(stderr, "+----+----+----+\n");
  fprintf(stderr, "|    | -0 | +0 |\n");
  fprintf(stderr, "+----+----+----+\n");
  fprintf(stderr, "| -0 | %2d | %2d |\n", m, p);
  fprintf(stderr, "+----+----+----+\n");
}

void prt_1_double(char *title, double m, double p)
{
  fprintf(stderr, "\n");
  fprintf(stderr, "%s\n", title);
  fprintf(stderr, "+----+------+------+\n");
  fprintf(stderr, "|    |  -0  |  +0  |\n");
  fprintf(stderr, "+----+------+------+\n");
  fprintf(stderr, "| -0 | %4.0f | %4.0f |\n", m, p);
  fprintf(stderr, "+----+------+------+\n");
}

void prt_2_bool_t(char *title, int mm, int mp, int pm, int pp)
{
  fprintf(stderr, "\n");
  fprintf(stderr, "%s\n", title);
  fprintf(stderr, "+----+----+----+\n");
  fprintf(stderr, "|    | -0 | +0 |\n");
  fprintf(stderr, "+----+----+----+\n");
  fprintf(stderr, "| -0 | %2d | %2d |\n", mm, mp);
  fprintf(stderr, "+----+----+----+\n");
  fprintf(stderr, "| +0 | %2d | %2d |\n", pm, pp);
  fprintf(stderr, "+----+----+----+\n");
}

void prt_2_double(char *title, double mm, double mp, double pm, double pp)
{
  fprintf(stderr, "\n");
  fprintf(stderr, "%s\n", title);
  fprintf(stderr, "+----+------+------+\n");
  fprintf(stderr, "|    |  -0  |  +0  |\n");
  fprintf(stderr, "+----+------+------+\n");
  fprintf(stderr, "| -0 | %4.0f | %4.0f |\n", mm, mp);
  fprintf(stderr, "+----+------+------+\n");
  fprintf(stderr, "| +0 | %4.0f | %4.0f |\n", pm, pp);
  fprintf(stderr, "+----+------+------+\n");
}

