/* r6test --- test program for r6.h, r6x6.h  */
/* Last edited on 2007-01-14 02:25:56 by stolfi */

#include <r6.h>
#include <r6x6.h>

#include <affirm.h>
#include <jsrandom.h>
#include <flt.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define N 6

void check_eq(double x, double y, char *msg);
void check_i_eq(int i, double x, double y, char *msg);
void check_ij_eq(int i, int j, double x, double y, char *msg);
void check_eps(double x, double y, double eps, char *msg);

void check_eq(double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** %20.16e %20.16e\n", (x), (y));
        affirm(0, msg);
      }
  }

void check_i_eq(int i, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d] %20.16e %20.16e\n", (i), (x), (y));
        affirm(0, msg);
      }
  }

void check_ij_eq(int i, int j, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d,%d] %20.16e %20.16e\n", (i), (j), (x), (y));
        affirm(0, msg);
      }
  }

void check_eps(double x, double y, double eps, char *msg)
  { if (fabs(x - y) > eps)
      { fprintf(stderr, " ** %20.16e %20.16e %20.16e\n", (x), (y), (eps));
        affirm(0, msg);
      }
  }

/* Internal prototypes */

int main (int argc, char **argv);
void test_r6(int verbose);
void test_r6x6(int verbose);
void throw_matrix(r6x6_t *m);
double frac (double x);

int main (int argc, char **argv)
  { int i;
    srand(1993);
    srandom(1993);

    for (i = 0; i < 100; i++) test_r6(i < 3);
    for (i = 0; i < 100; i++) test_r6x6(i < 3);
    fclose(stderr);
    fclose(stdout);
    return (0);
  }

double frac (double x)
  { int i = (int)x;
    double f = x - (double)i;
    return (f);
  }

void test_r6(int verbose)
  { r6_t a, b, c, d, e, f, g, para, perp;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;

    if (verbose)
      { fprintf(stderr,
          "sizeof(r6_t) = %d  %d*sizeof(double) = %d\n",
          sizeof(r6_t), N, N*sizeof(double)
        );
      }

    /* r6_zero: */
    r6_zero(&a);
    for (i = 0; i < N; i++)
      { check_eq(a.c[i], 0.0, "r6_zero error"); }

    /* r6_all: */
    r6_all(3.14, &a);
    for (i = 0; i < N; i++)
      { check_eq(a.c[i], 3.14, "r6_all error"); }

    /* r6_axis: */
    for (k = 0; k < N; k++)
      { r6_axis(k, &a);
        for (i = 0; i < N; i++)
          { check_eq(a.c[i], (i == k ? 1.0 : 0.0), "r6_axis error"); }
      }

    /* r6_throw_cube */
    r6_throw_cube(&a);
    for (i = 0; i < N; i++)
      { affirm(a.c[i] != a.c[(i+1)%N], "r6_throw error(1)"); 
        affirm(frac(a.c[i]*256.0) != 0.0, "r6_throw error(1)"); 
        affirm((a.c[i] > -1.0) && (a.c[i] < 1.0), "r6_throw error(2)"); 
      }

    /* r6_add: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r6_add(&a, &b, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], a.c[i] + b.c[i], "r6_add error"); }

    /* r6_sub: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r6_sub(&a, &b, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], a.c[i] - b.c[i], "r6_sub error"); }

    /* r6_neg: */
    r6_throw_cube(&a);
    r6_neg(&a, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], - a.c[i], "r6_neg error"); }

    /* r6_scale: */
    s = drandom();
    r6_throw_cube(&a);
    r6_scale(s, &a, &d);
    for (i = 0; i < N; i++)
      { double zi = s*a.c[i];
        check_eq(d.c[i], zi, "r6_scale error(1)");
      }

    /* r6_mix: */
    s = drandom();
    t = drandom();
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r6_mix(s, &a, t, &b, &d);
    for (i = 0; i < N; i++)
      { double ddi = s * a.c[i] + t * b.c[i];
        check_eq(d.c[i], ddi, "r6_mix error");
      }

    /* r6_mix_in: */
    s = drandom();
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    d = b;
    r6_mix_in(s, &a, &d);
    for (i = 0; i < N; i++)
      { double ddi = b.c[i] + s * a.c[i];
        check_eq(d.c[i], ddi, "r6_mix_in error");
      }

    /* r6_weigh: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r6_weigh(&a, &b, &d);
    for (i = 0; i < N; i++)
      { double ddi = a.c[i] * b.c[i];
        check_eq(d.c[i], ddi, "r6_weigh error");
      }

    /* r6_norm, r6_norm_sqr, r6_L1_inf_norm: */
    r6_throw_cube(&a);
    r = r6_norm(&a);
    s = r6_norm_sqr(&a);
    t = r6_L_inf_norm(&a);
    ss = 0.0;
    tt = 0.0;
    for (i = 0; i < N; i++)
      { double ai = fabs(a.c[i]);
        ss += ai*ai; 
        if (ai > tt) { tt = ai; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "r6_norm error");
    check_eps(s, ss, 0.000000001 * ss, "r6_norm_sqr error");
    check_eq(t, tt, "r6_L_inf_norm error");

    /* r6_dist, r6_dist_sqr, r6_L_inf_dist: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r = r6_dist(&a, &b);
    s = r6_dist_sqr(&a, &b);
    t = r6_L_inf_dist(&a, &b);
    ss = 0.0;
    tt = 0.0;

    for (i = 0; i < N; i++)
      { double di = fabs(a.c[i] - b.c[i]);
        ss += di*di; 
        if (di > tt) { tt = di; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "r6_dist error");
    check_eps(s, ss, 0.000000001 * ss, "r6_dist_sqr error");
    check_eq(t, tt, "r6_L_inf_dist error");

    /* r6_dir, r6_L_inf_dir: */
    r6_throw_cube(&a);
    r = r6_dir(&a, &b);
    s = r6_L_inf_dir(&a, &d);
    ss = r6_norm(&a);
    tt = r6_L_inf_norm(&a);
    for (i = 0; i < N; i++)
      { check_eps(b.c[i], a.c[i]/ss, 0.000000001 * ss, "r6_dir error");
        check_eps(d.c[i], a.c[i]/tt, 0.000000001 * tt, "r6_L_inf_dir error");
      }

    /* r6_dot, r6_cos, r6_sin: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r = r6_dot(&a, &b);
    s = r6_sin(&a, &b);
    t = r6_cos(&a, &b);
    mag = sqrt(r6_dot(&a,&a)*r6_dot(&b,&b));
    rr = 0.0;
    for (i = 0; i < N; i++) { rr += a.c[i]*b.c[i]; }
    tt = rr/(r6_norm(&a)*r6_norm(&b));
    check_eps(r, rr, 0.000000001 * mag, "r6_dot error(1)");
    check_eps(t, tt, 0.000000001, "r6_cos error(1)");
    d = a;
    r6_mix_in(-rr/r6_norm_sqr(&b), &b, &d);
    ss = r6_norm(&d)/r6_norm(&a);
    check_eps(s, ss, 0.000000001, "r6_sin error(1)");
    for (i = 0; i < N; i++)
      { r6_axis(i, &a);
        for (j = 0; j < N; j++)
          { r6_axis(j, &b);
            r = r6_dot(&a, &b);
            s = r6_sin(&a, &b);
            t = r6_cos(&a, &b);
            rr = (i == j ? 1.0 : 0.0);
            check_eq(r, rr, "r6_dot error(2)");
            check_eq(t, rr, "r6_dot error(2)");
            check_eq(s, 1.0 - rr, "r6_dot error(2)");
          }
      }

    /* r6_cross: */
    /* Test on basis vectors: */
    for (i = 0; i < N; i++)
      { int i0 = (i + 0) % N;
        int i1 = (i + 1) % N;
        int i2 = (i + 2) % N;
        int i3 = (i + 3) % N;
        int i4 = (i + 4) % N;
        int i5 = (i + 5) % N;
        double sign = ((i % 2) == 0 ? 1.0 : -1.0);
        int p;
        r6_axis(i0, &a);
        r6_axis(i1, &b);
        r6_axis(i2, &c);
        r6_axis(i3, &d);
        r6_axis(i4, &e);
        r6_cross(&a, &b, &c, &d, &e, &f);
        r6_axis(i5, &g);
        for (p = 0; p < N; p++)
          { double gp = sign*g.c[p];
            check_eq(f.c[p], gp, "r6_cross error(x)");
          }
      }
    /* Test on random vectors: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r6_throw_cube(&c);
    r6_throw_cube(&d);
    r6_throw_cube(&e);
    r6_cross(&a, &b, &c, &d, &e, &f);
    mag = r6_norm(&a)*r6_norm(&b)*r6_norm(&c)*r6_norm(&d)*r6_norm(&e);
    r = r6_dot(&a, &f);
    check_eps(r, 0.0, 0.00000001 * mag*r6_norm(&a), "r6_cross error(1)");
    r = r6_dot(&b, &f);
    check_eps(r, 0.0, 0.00000001 * mag*r6_norm(&b), "r6_cross error(2)");
    r = r6_dot(&c, &f);
    check_eps(r, 0.0, 0.00000001 * mag*r6_norm(&c), "r6_cross error(3)");
    r = r6_dot(&d, &f);
    check_eps(r, 0.0, 0.00000001 * mag*r6_norm(&d), "r6_cross error(3)");
    r = r6_dot(&e, &f);
    check_eps(r, 0.0, 0.00000001 * mag*r6_norm(&e), "r6_cross error(3)");

    /* r6_det: */
    /* Test on basis vectors: */
    for (i = 0; i < N; i++)
      { int i0 = (i + 0) % N;
        int i1 = (i + 1) % N;
        int i2 = (i + 2) % N;
        int i3 = (i + 3) % N;
        int i4 = (i + 4) % N;
        int i5 = (i + 5) % N;
        double sign = ((i % 2) == 0 ? 1.0 : -1.0);
        r6_axis(i0, &a);
        r6_axis(i1, &b);
        r6_axis(i2, &c);
        r6_axis(i3, &d);
        r6_axis(i4, &e);
        r6_axis(i5, &f);
        r = r6_det(&a, &b, &c, &d, &e, &f);
        check_eq(r, sign, "r6_det error(2)");
      }
    /* Test on random vectors: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r6_throw_cube(&c);
    r6_throw_cube(&d);
    r6_throw_cube(&e);
    r6_throw_cube(&f);
    r = r6_det(&a, &b, &c, &d, &e, &f);
    r6_cross(&a, &b, &c, &d, &e, &g);
    rr = r6_dot(&f, &g);
    mag = r6_norm(&a)*r6_norm(&b)*r6_norm(&c)*r6_norm(&d)*r6_norm(&e)*r6_norm(&f);
    check_eps(r, rr, 0.00000001 * mag, "r6_det error(1)");

    /* r6_decomp: */
    r6_throw_cube(&a);
    r6_throw_cube(&b);
    r = r6_decomp(&a, &b, &para, &perp);
    rr = r6_dot(&a, &b)/r6_norm_sqr(&b);  
    check_eps(r, rr, 0.000000001 * (fabs(r) + fabs(rr)), "r6_decomp error(1)");
    r6_add(&para, &perp, &c);
    s = r6_dist(&a, &c);
    affirm (s <= 0.000000001 * r6_norm(&a), "r6_decomp error(2)");
    s = r6_dot(&perp, &b);
    check_eps(s, 0.0, 0.000000001 * r6_norm(&b), "r6_decomp error(3)");
    t = r6_dot(&para, &perp);
    check_eps(t, 0.0, 0.000000001 * r6_norm(&a), "r6_decomp error(4)");

    /* r6_print: */
    if (verbose)
      { r6_throw_cube (&a);
        fprintf(stderr, "a = ");
        r6_print(stderr, &a);
        fputc('\n', stderr);
      }
  }

void test_r6x6(int verbose)
  { r6x6_t A, B, C;
    r6_t a, b, c, bb, cc;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;

    /* Size and allocation: */
    if (verbose)
      { fprintf(stderr,
          "sizeof(r6x6_t) = %d  %d*%d*sizeof(double) = %d\n",
          sizeof(r6x6_t), N, N, N*N*sizeof(double)
        );
        fprintf(stderr, "&B = %08x\n", (unsigned)&B);
        fprintf(stderr, "&A-&B = %d\n", ((unsigned)(&A))-((unsigned)(&B)));
        fprintf(stderr, "&B-&C = %d\n", ((unsigned)(&B))-((unsigned)(&C)));
        fprintf(stderr, "&(B.c) = %08x\n", (unsigned)&(B.c));
        fprintf(stderr, "B.c = %08x\n", (unsigned)(B.c));
        fprintf(stderr, "&(B.c[0]) = %08x\n", (unsigned)&(B.c[0]));
        fprintf(stderr, "B.c[0] = %08x\n", (unsigned)(B.c[0]));
        fprintf(stderr, "&(B.c[0][0]) = %08x\n", (unsigned)&(B.c[0][0]));
      }

    /* Indexing and addressing: */
    for (i = 0; i < N; i++)
      for (j = 0; j < N; j++)
        { double *Bij = &(B.c[i][j]); 
          affirm(Bij = ((double *)&B)+(N*i)+j, "r6x6_t indexing error");
        }

    /* r6x6_zero, r6x6_ident: */
    r6x6_zero(&A);
    r6x6_ident(&B);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { check_eq(A.c[i][j], 0.0, "r6x6_zero error"); 
            check_eq(B.c[i][j], (i == j ? 1.0 : 0.0), "r6x6_ident error");
          }
      }

    /* r6x6_map_row, r6x6_map_col: */
    throw_matrix(&A);
    r6_throw_cube(&a);
    r6x6_map_row(&a, &A, &b);
    r6x6_map_col(&A, &a, &c);
    r6_zero(&bb);
    r6_zero(&cc);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { bb.c[j] += a.c[i] * A.c[i][j];
            cc.c[i] += A.c[i][j] * a.c[j];
          }
      }
    r = r6_dist(&b, &bb);
    affirm(r < 0.000000001 * r6_norm(&bb), "r6_map_row error");
    s = r6_dist(&c, &cc);
    affirm(s < 0.000000001 * r6_norm(&cc), "r6_map_col error");

    /* r6x6_mul: */
    throw_matrix(&A);
    throw_matrix(&B);
    r6x6_mul(&A, &B, &C);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { double sum = 0.0;
            for (k = 0; k < N; k++) { sum += A.c[i][k]*B.c[k][j]; }
            check_eps(C.c[i][j], sum, 0.000000001 * fabs(sum),
              "r6x6_mul error"
            );
          }
      }

    /* r6x6_det: */
    throw_matrix(&A);
    for (i = 0; i < N; i++)
      { int k = (i + 1) % N;
        for (j = 0; j < N; j++)
          { /* Check for linearity */
            r = drandom();
            A.c[i][j] = r;
            rr = r6x6_det(&A);

            s = drandom();
            A.c[i][j] = s;
            ss = r6x6_det(&A);

            t = drandom();
            A.c[i][j] = r*(1-t) + s*t;
            tt = r6x6_det(&A);
            mag = fabs(rr) + fabs(ss) + fabs(tt);
            check_eps(tt, (rr*(1 - t) + ss*t), 000000001 * mag,
              "r6x6_det error(1)"
            );
          }

        /* Row swap test: */
        r = r6x6_det(&A);
        for (j = 0; j < N; j++)
          { double t = A.c[i][j]; A.c[i][j] = A.c[k][j]; A.c[k][j] = t; }
        rr = r6x6_det(&A);
        mag = fabs(r) + fabs(rr);
        check_eps(r, -rr, 000000001 * mag, "r6x6_det error(2)");

        /* Col swap test: */
        r = r6x6_det(&A);
        for (j = 0; j < N; j++)
          { double t = A.c[j][i]; A.c[j][i] = A.c[j][k]; A.c[j][k] = t; }
        rr = r6x6_det(&A);
        mag = fabs(r) + fabs(rr);
        check_eps(r, -rr, 000000001 * mag, "r6x6_det error(3)");
      }

    /* r6x6_inv: */
    throw_matrix(&A);
    r6x6_inv(&A, &B);
    r6x6_mul(&A, &B, &C);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { double val = (i == j ? 1.0 : 0.0);
            affirm((C.c[i][j] - val) < 000000001, "r6x6_inv error");
          }
      }

    /* r6x6_print: */
    if (verbose)
      { throw_matrix (&A);
        fprintf(stderr, "A = ");
        r6x6_print(stderr, &A);
        fputc('\n', stderr);
      }
  }  

void throw_matrix(r6x6_t *m)
  { int i, j;
    r6_t a;
    for (i = 0; i < N; i++)
      { r6_throw_cube(&a);
        for (j = 0; j < N; j++) { m->c[i][j] = a.c[j]; }
      }
  }

