/* rntest --- test program for rn.h, rmxn.h  */
/* Last edited on 2007-02-09 11:35:19 by stolfi */

#include <rn.h>
#include <rmxn.h>
#include <rmxn_extra.h>

#include <affirm.h>
#include <jsrandom.h>
#include <jsmath.h>
#include <flt.h>
#include <bool.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <values.h>

void check_eq(double x, double y, char *msg);
void check_i_eq(int i, double x, double y, char *msg);
void check_ij_eq(int i, int j, double x, double y, char *msg);
void check_eps(double x, double y, double eps, char *msg);

void check_eq(double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** %20.16e %20.16e\n", (x), (y));
        affirm(0, msg);
      }
  }

void check_i_eq(int i, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d] %20.16e %20.16e\n", (i), (x), (y));
        affirm(0, msg);
      }
  }

void check_ij_eq(int i, int j, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d,%d] %20.16e %20.16e\n", (i), (j), (x), (y));
        affirm(0, msg);
      }
  }

void check_eps(double x, double y, double eps, char *msg)
  { if (fabs(x - y) > eps)
      { fprintf(stderr, " ** %20.16e %20.16e %20.16e\n", (x), (y), (eps));
        affirm(0, msg);
      }
  }
  
/* Internal prototypes */

int main (int argc, char **argv);
void test_rn(int verbose);
void test_rmxn(int verbose);
void throw_matrix(int m, int n, double *Amn);
void throw_LT_matrix(int m, double *Lmm);
double frac (double x);

void check_simplex(int n, double V[]);
  /* Checks whether the simplex {V} is regular (modulo some roundoff).
    Assumes that {V} has {n+1} rows (vertices) and {n} columns
    (coordinates). */

int main (int argc, char **argv)
  { int i;
    srand(1993);
    srandom(1993);
    for (i = 0; i < 100; i++) test_rn(i < 3);
    for (i = 0; i < 100; i++) test_rmxn(i < 3);
    fclose(stderr);
    fclose(stdout);
    return (0);
  }

double frac (double x)
  { int i = (int)x;
    double f = x - (double)i;
    return (f);
  }

void test_rn (int verbose)
  { double *a, *b, *c, *d, *e, *para, *perp;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;
    int maxsize = (verbose ? 5 : 10);
    int n = rand()/(RAND_MAX/maxsize) + 1;  /* Vector size. */

    fprintf(stderr, "test_rn:  n = %d\n", n);
    
    a = rn_alloc(n);
    b = rn_alloc(n);
    c = rn_alloc(n);
    d = rn_alloc(n);
    e = rn_alloc(n);
    para = rn_alloc(n);
    perp = rn_alloc(n);

    /* rn_zero: */
    rn_zero(n, a);
    for (i = 0; i < n; i++)
      { check_i_eq(i, a[i], 0.0, "rn_zero error"); }

    /* rn_all: */
    rn_all(n, 3.14, a);
    for (i = 0; i < n; i++)
      { check_i_eq(i, a[i], 3.14, "rn_all error"); }

    /* rn_axis: */
    for (k = 0; k < n; k++)
      { rn_axis(n, k, a);
        for (i = 0; i < n; i++)
          { check_i_eq(i, a[i], (i == k ? 1.0 : 0.0), "rn_axis error"); }
      }

    /* rn_throw_cube */
    rn_throw_cube(n, a);
    for (i = 0; i < n; i++)
      { affirm((n == 1) || (a[i] != a[(i+1)%n]), "rn_throw error(1)"); 
        affirm(frac(a[i]*256.0) != 0.0, "rn_throw error(1)"); 
        affirm((a[i] > -1.0) && (a[i] < 1.0), "rn_throw error(2)"); 
      }

    /* rn_add: */
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    rn_add(n, a, b, d);
    for (i = 0; i < n; i++)
      { check_i_eq(i, d[i], a[i] + b[i], "rn_add error"); }

    /* rn_sub: */
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    rn_sub(n, a, b, d);
    for (i = 0; i < n; i++)
      { check_i_eq(i, d[i], a[i] - b[i], "rn_sub error"); }

    /* rn_neg: */
    rn_throw_cube(n, a);
    rn_neg(n, a, d);
    for (i = 0; i < n; i++)
      { double dxi = -a[i];
        check_i_eq(i, d[i], dxi, "rn_neg error"); }

    /* rn_scale: */
    s = drandom();
    rn_throw_cube(n, a);
    rn_scale(n, s, a, d);
    for (i = 0; i < n; i++)
      { double dxi = s*a[i];
        fprintf(stderr, "%5d %24.16e %24.16e %24.16e %24.16e\n", i, s, a[i], d[i], dxi);
        check_i_eq(i, d[i], dxi, "rn_scale error(1)");
      }

    /* rn_mix: */
    s = drandom();
    t = drandom();
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    rn_mix(n, s, a, t, b, d);
    for (i = 0; i < n; i++)
      { double dxi = s * a[i] + t * b[i];
        check_i_eq(i, d[i], dxi, "rn_mix error");
      }

    /* rn_mix_in: */
    s = drandom();
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    for (i = 0; i < n; i++) { c[i] = b[i]; }
    rn_mix_in(n, s, a, c);
    for (i = 0; i < n; i++)
      { double dxi = b[i] + s * a[i];
        check_i_eq(i, c[i], dxi, "rn_mix_in error");
      }

    /* rn_weigh: */
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    rn_weigh(n, a, b, d);
    for (i = 0; i < n; i++)
      { double dxi = a[i] * b[i];
        check_i_eq(i, d[i], dxi, "rn_weigh error");
      }

    /* rn_norm, rn_norm_sqr, rn_L1_inf_norm: */
    rn_throw_cube(n, a);
    r = rn_norm(n, a);
    s = rn_norm_sqr(n, a);
    t = rn_L_inf_norm(n, a);
    ss = 0.0;
    tt = 0.0;
    for (i = 0; i < n; i++)
      { double ai = fabs(a[i]);
        ss += ai*ai; 
        if (ai > tt) { tt = ai; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "rn_norm error");
    check_eps(s, ss, 0.000000001 * ss, "rn_norm_sqr error");
    check_eq(t, tt, "rn_L_inf_norm error");

    /* rn_dist, rn_dist_sqr, rn_L_inf_dist: */
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    r = rn_dist(n, a, b);
    s = rn_dist_sqr(n, a, b);
    t = rn_L_inf_dist(n, a, b);

    ss = 0.0;
    tt = 0.0;
    for (i = 0; i < n; i++)
      { double di = fabs(a[i] - b[i]);
        ss += di*di; 
        if (di > tt) { tt = di; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "rn_dist error");
    check_eps(s, ss, 0.000000001 * ss, "rn_dist_sqr error");
    check_eq(t, tt, "rn_L_inf_dist error");

    /* rn_dir, rn_L_inf_dir: */
    rn_throw_cube(n, a);
    r = rn_dir(n, a, b);
    s = rn_L_inf_dir(n, a, d);
    ss = rn_norm(n, a);
    tt = rn_L_inf_norm(n, a);
    for (i = 0; i < n; i++)
      { check_eps(b[i], a[i]/ss, 0.000000001 * ss, "rn_dir error");
        check_eps(d[i], a[i]/tt, 0.000000001 * tt, "rn_L_inf_dir error");
      }

    /* rn_dot, rn_cos, rn_sin: */
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    r = rn_dot(n, a, b);
    s = rn_sin(n, a, b);
    t = rn_cos(n, a, b);
    mag = sqrt(rn_dot(n, a,a)*rn_dot(n, b,b));
    rr = 0.0;
    for (i = 0; i < n; i++) { rr += a[i]*b[i]; }
    tt = rr/(rn_norm(n, a)*rn_norm(n, b));
    check_eps(r, rr, 0.000000001 * mag, "rn_dot error(1)");
    check_eps(t, tt, 0.000000001, "rn_cos error(1)");
    for (i = 0; i < n; i++) { d[i] = a[i]; }
    rn_mix_in(n, -rr/rn_norm_sqr(n, b), b, d);
    ss = rn_norm(n, d)/rn_norm(n, a);
    check_eps(s, ss, 0.000000001, "rn_sin error(1)");
    for (i = 0; i < n; i++)
      { rn_axis(n, i, a);
        for (j = 0; j < n; j++)
          { rn_axis(n, j, b);
            r = rn_dot(n, a, b);
            s = rn_sin(n, a, b);
            t = rn_cos(n, a, b);
            rr = (i == j ? 1.0 : 0.0);
            check_ij_eq(i, j, r, rr, "rn_dot error(2)");
            check_ij_eq(i, j, t, rr, "rn_dot error(3)");
            check_ij_eq(i, j, s, 1.0 - rr, "rn_dot error(4)");
          }
      }

    /* rn_cross: */
    { double **z = (double **)malloc((n-1)*sizeof(double *));
      double *magz = rn_alloc(n);
      /* Test on basis vectors: */
      for (i = 0; i < n; i++)
        { double sign = ((n-1)*i % 2 == 0 ? +1.0 : -1.0);
          for (k = 0; k < n-1; k++)
            { int ik = (i + k) % n; 
              z[k] = rn_alloc(n);
              rn_axis(n, ik, z[k]);
            }
          { int in1 = (i + n-1) % n; rn_axis(n, in1, b); }
          rn_cross(n, z, c);
          for (j = 0; j < n; j++)
            { double cxj = sign*b[j];
              check_ij_eq(i, j, c[j], cxj, "rn_cross error(x)");
            }
        }
      /* Test on random vectors: */
      mag = 1.0;
      for (k = 0; k < n-1; k++)
        { z[k] = rn_alloc(n);
          rn_throw_cube(n, z[k]);
          { magz[k] = rn_norm(n, z[k]); mag *= magz[k]; }
        }
      rn_cross(n, z, c);
      for (k = 0; k < n-1; k++)
        { r = rn_dot(n, z[k], c);
          check_eps(r, 0.0, 0.00000001 * mag*magz[k], "rn_cross error(1)");
        }
      for (k = 0; k < n-1; k++) { free(z[k]); }
      free(z); free(magz);
    }
    
    /* rn_det: */
    { double **z = (double **)malloc(n*sizeof(double *));
      double *magz = rn_alloc(n);
      /* Test on basis vectors: */
      for (i = 0; i < n; i++)
        { double sign = ((n-1)*i % 2 == 0 ? +1.0 : -1.0);
          for (k = 0; k < n; k++)
            { int ik = (i + k) % n; 
              z[k] = rn_alloc(n);
              rn_axis(n, ik, z[k]);
            }
          r = rn_det(n, z);
          check_i_eq(i, r, sign, "rn_det error(2)");
        }
      /* Test on random vectors: */
      mag = 1.0;
      for (k = 0; k < n; k++)
        { z[k] = rn_alloc(n);
          rn_throw_cube(n, z[k]);
          { magz[k] = rn_norm(n, z[k]); mag *= magz[k]; }
        }
      r = rn_det(n, z);
      rn_cross(n, z, c);
      rr = rn_dot(n, c, z[n-1]);
      check_eps(r, rr, 0.00000001 * mag, "rn_det error(1)");

      for (k = 0; k < n; k++) { free(z[k]); }
      free(z); free(magz);
    }

    /* rn_decomp: */
    rn_throw_cube(n, a);
    rn_throw_cube(n, b);
    r = rn_decomp(n, a, b, para, perp);
    rr = rn_dot(n, a, b)/rn_norm_sqr(n, b);  
    check_eps(r, rr, 0.000000001 * (fabs(r) + fabs(rr)), "rn_decomp error(1)");
    rn_add(n, para, perp, c);
    s = rn_dist(n, a, c);
    check_eps(s, 0.0, 0.000000001 * rn_norm(n, a), "rn_decomp error(2)");
    s = rn_dot(n, perp, b);
    check_eps(s, 0.0, 0.000000001 * rn_norm(n, b), "rn_decomp error(3)");
    t = rn_dot(n, para, perp);
    check_eps(t, 0.0, 0.000000001 * rn_norm(n, a), "rn_decomp error(4)");

    /* rn_print: */
    if (verbose)
      { rn_throw_cube (n, a);
        fprintf(stderr, "a = ");
        rn_print(stderr, n, a);
        fputc('\n', stderr);
      }
      
    free(a);
    free(b);
    free(c);
    free(d);
    free(e);
    free(para);
    free(perp);
  }

void test_rmxn(int verbose)
  { double *Amn, *Bmn, *Cmn, *Amp, *Bpn, *Amm, *Bmm, *Cmm;
    double *am, *bm, *cm, *an, *bn, *cn;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;
    int maxsize = (verbose ? 5 : 10);
    int m = rand()/(RAND_MAX/maxsize) + 1;  /* Number of rows. */
    int n = rand()/(RAND_MAX/maxsize) + 1;  /* Number of columns. */
    int p = rand()/(RAND_MAX/maxsize) + 1;  /* Middle dimension for rmxn_mul. */

    fprintf(stderr, "test_rmxn:  m = %d  n = %d  p = %d\n", m, n, p);
    
    Amn = rmxn_alloc(m, n);
    Bmn = rmxn_alloc(m, n);
    Cmn = rmxn_alloc(m, n);

    Amp = rmxn_alloc(m, p);
    Bpn = rmxn_alloc(p, n);

    Amm = rmxn_alloc(m, m);
    Bmm = rmxn_alloc(m, m);
    Cmm = rmxn_alloc(m, m);

    am = rn_alloc(m);
    bm = rn_alloc(m);
    cm = rn_alloc(m);
    
    an = rn_alloc(n);
    bn = rn_alloc(n);
    cn = rn_alloc(n);

    /* rmxn_zero, rmxn_ident: */
    rmxn_zero(m, n, Amn);
    rmxn_ident(m, n, Bmn);
    for (i = 0; i < m; i++)
      { for (j = 0; j < n; j++)
          { check_ij_eq(i, j, Amn[n*i + j], 0.0, "rmxn_zero error"); 
            check_ij_eq(i, j, Bmn[n*i + j], (i == j ? 1.0 : 0.0), "rmxn_ident error");
          }
      }

    /* rmxn_scale: */
    s = drandom();
    throw_matrix(m, n, Amn);
    rmxn_scale(m, n, s, Amn, Cmn);
    for (i = 0; i < m; i++)
      { for (j = 0; j < n; j++)
          { double zij = s*Amn[n*i + j];
            check_ij_eq(i, j, Cmn[n*i + j], zij, "rmxn_scale error(1)");
          }
      }

    /* rmxn_mix: */
    s = drandom();
    t = drandom();
    throw_matrix(m, n, Amn);
    throw_matrix(m, n, Bmn);
    rmxn_mix(m, n, s, Amn, t, Bmn, Cmn);
    for (i = 0; i < m; i++)
      { for (j = 0; j < n; j++)
          { double zij = s*Amn[n*i + j] + t*Bmn[n*i + j];
            check_ij_eq(i, j, Cmn[n*i + j], zij, "rmxn_mix error(1)");
          }
      }

    /* rmxn_rel_diff: */
    s = drandom();
    t = drandom();
    throw_matrix(m, n, Amn);
    throw_matrix(m, n, Bmn);
    rmxn_rel_diff(m, n, Amn, Bmn, Cmn);
    for (i = 0; i < m; i++)
      { for (j = 0; j < n; j++)
          { double dij = rel_diff(Amn[n*i + j], Bmn[n*i + j]);
            check_ij_eq(i, j, Cmn[n*i + j], dij, "rmxn_rel_diff error(1)");
          }
      }

    /* rmxn_map_row, rmxn_map_col: */
    throw_matrix(m, n, Amn);
    rn_throw_cube(m, am);
    rn_throw_cube(n, an);
    rmxn_map_row(m, n, am, Amn, bn);
    rmxn_map_col(m, n, Amn, an, bm);
    rn_zero(m, cm);
    rn_zero(n, cn);
    for (i = 0; i < m; i++)
      { for (j = 0; j < n; j++)
          { cn[j] += am[i] * Amn[n*i + j];
            cm[i] += Amn[n*i + j] * an[j];
          }
      }
    r = rn_dist(m, bm, cm);
    check_eps(r, 0.0, 0.000000001 * rn_norm(m, cm), "rn_map_row error");
    s = rn_dist(n, bn, cn);
    check_eps(s, 0.0, 0.000000001 * rn_norm(n, cn), "rn_map_col error");

    /* rmxn_mul: */
    throw_matrix(m, p, Amp);
    throw_matrix(p, n, Bpn);
    rmxn_mul(m, p, n, Amp, Bpn, Cmn);
    for (i = 0; i < m; i++)
      { for (j = 0; j < n; j++)
          { double sum = 0.0;
            for (k = 0; k < p; k++) { sum += Amp[p*i + k]*Bpn[n*k + j]; }
            check_eps(Cmn[n*i + j], sum, 0.000000001 * fabs(sum), 
              "rmxn_mul error"
            );
          }
      }

    /* rmxn_mul_tr: */
    throw_matrix(m, n, Amn);
    throw_matrix(p, n, Bpn);
    rmxn_mul_tr(m, p, n, Amn, Bpn, Amp);
    for (i = 0; i < m; i++)
      { for (j = 0; j < p; j++)
          { double sum = 0.0;
            for (k = 0; k < n; k++) { sum += Amn[n*i + k]*Bpn[n*j + k]; }
            check_eps(Amp[p*i + j], sum, 0.000000001 * fabs(sum), 
              "rmxn_mul_tr error"
            );
          }
      }

    /* rmxn_det: */
    throw_matrix(m, m, Amm);
    for (i = 0; i < m; i++)
      { int k = (i + 1) % m;
        for (j = 0; j < m; j++)
          { /* Check for linearity */
            r = drandom();
            Amm[m*i + j] = r;
            rr = rmxn_det(m, Amm);

            s = drandom();
            Amm[m*i + j] = s;
            ss = rmxn_det(m, Amm);

            t = drandom();
            Amm[m*i + j] = r*(1-t) + s*t;
            tt = rmxn_det(m, Amm);
            mag = fabs(rr) + fabs(ss) + fabs(tt);
            check_eps(tt, (rr*(1-t) + ss*t), 000000001 * mag, 
              "rmxn_det error(1)"
            );
          }

        /* Row swap test: */
        r = rmxn_det(m, Amm);
        for (j = 0; j < m; j++)
          { double *Aij = &(Amm[m*i + j]);
            double *Akj = &(Amm[m*k + j]);
            { double t = *Aij; *Aij = *Akj; *Akj = t; }
          }
        rr = rmxn_det(m, Amm);
        mag = fabs(r) + fabs(rr);
        check_eps(r, (-rr), 000000001 * mag, "rmxn_det error(2)");

        /* Col swap test: */
        r = rmxn_det(m, Amm);
        for (j = 0; j < m; j++)
          { double *Aji = &(Amm[m*j + i]);
            double *Ajk = &(Amm[m*j + k]);
            { double t = *Aji; *Aji = *Ajk; *Ajk = t; }
          }
        rr = rmxn_det(m, Amm);
        mag = fabs(r) + fabs(rr);
        check_eps(r, (-rr), 000000001 * mag, "rmxn_det error(3)");
      }

    /* rmxn_cof: */
    if (verbose) { fprintf(stderr, "** warning: rmxn_cof not tested\n"); }
    
    /* rmxn_adj: */
    if (verbose) { fprintf(stderr, "** warning: rmxn_adj not tested\n"); }
    
    /* rmxn_inv: */
    throw_matrix(m, m, Amm);
    rmxn_inv(m, Amm, Bmm);
    rmxn_mul(m, m, m, Amm, Bmm, Cmm);
    for (i = 0; i < m; i++)
      { for (j = 0; j < m; j++)
          { double val = (i == j ? 1.0 : 0.0);
            check_eps(Cmm[m*i + j], val, 0.000000001, "rmxn_inv error");
          }
      }

    /* rmxn_cholesky: */
    { double *Lmm = Amm, *Jmm = Cmm;
      throw_LT_matrix(m, Jmm);
      for(i = 0; i < m; i++) { Jmm[m*i + i] = fabs(Jmm[m*i + i]) + 1.0; }
      rmxn_mul_tr(m, m, m, Jmm, Jmm, Bmm);
      rmxn_cholesky(m, Bmm, Lmm);
      for (i = 0; i < m; i++)
        { for (j = 0; j < m; j++)
            { double eps = 0.000000001;
              check_eps(Lmm[m*i + j], Jmm[m*i + j], eps, 
                "rmxn_cholesky error"
              );
            }
        }
    }
      
    /* rmxn_LT_inv_map_row, rmxn_LT_inv_map_col: */
    { double *Lmm = Amm;
      int by_row; 
      throw_LT_matrix(m, Lmm);
      rmxn_inv(m, Lmm, Bmm);
      rn_throw_cube(m, am);
      for (by_row = 0; by_row < 2; by_row++)
        { char *msg;
          if (by_row) 
            { rmxn_LT_inv_map_row(m, am, Lmm, bm);
              rmxn_map_row(m, m, am, Bmm, cm);
              msg = "rn_LT_inv_map_row error";
            }
          else
            { rmxn_LT_inv_map_col(m, Lmm, am, bm);
              rmxn_map_col(m, m, Bmm, am, cm);
              msg = "rn_LT_inv_map_col error";
            }
          r = rn_dist(m, bm, cm);
          check_eps(r, 0.0, 0.000000001 * rn_norm(m, cm), msg);
        }
    }
      
    /* rmxn_LT_pos_div: */
    { double *Lmm = Amm, *Jmm = Bmm;
      /* We must use "Amn" as if it were "n x m" not "m x n". */
      double *Anm = Amn, *Bnm = Bmn, *Cnm = Cmn; 
      throw_LT_matrix(m, Lmm);
      rmxn_inv(m, Lmm, Jmm);
      throw_matrix(n, m, Anm);
      rmxn_LT_pos_div(n, m, Anm, Lmm, Cnm);
      rmxn_mul(n, m, m, Anm, Jmm, Bnm);
      for (i = 0; i < n; i++)
        { for (j = 0; j < m; j++)
            { double eps = 0.000000001*(fabs(Bnm[m*i + j]) + 1.0e-200);
              check_eps(Cnm[m*i + j], Bnm[m*i + j], eps, 
                "rmxn_LT_pos_div error"
              );
            }
        }
    }
    
    /* rmxn_LT_pre_div: */
    { double *Lmm = Amm, *Jmm = Bmm;
      throw_LT_matrix(m, Lmm);
      rmxn_inv(m, Lmm, Jmm);
      throw_matrix(m, n, Amn);
      rmxn_LT_pre_div(m, n, Lmm, Amn, Cmn);
      rmxn_mul(m, m, n, Jmm, Amn, Bmn);
      for (i = 0; i < m; i++)
        { for (j = 0; j < n; j++)
            { double eps = 0.000000001*fabs(Bmn[n*i + j]);
              check_eps(Cmn[n*i + j], Bmn[n*i + j], eps, 
                "rmxn_LT_pre_div error"
              );
            }
        }
    }
     
    /* rmxn_print: */
    if (verbose)
      { throw_matrix (m, n, Amn);
        fprintf(stderr, "Amn = ");
        rmxn_print(stderr, m, n, Amn);
        fputc('\n', stderr);
      }

    /* rmxn_full_simplex: */
    { double V[(n+1)*n];
      rmxn_full_simplex(n, V);
      check_simplex(n, V);
    }

    /* rmxn_spin_rows: */
    if (verbose) { fprintf(stderr, "** warning: rmxn_spin_rows not tested\n"); }
    
    /* rmxn_shift_rows: */
    if (verbose) { fprintf(stderr, "** warning: rmxn_shift_rows not tested\n"); }
    
    free(Amn); free(Bmn); free(Cmn);
    free(Amp); free(Bpn);
    free(Amm); free(Bmm); free(Cmm);
    free(am); free(bm); free(cm);
    free(an); free(bn); free(cn);
  }  

void throw_matrix(int m, int n, double *Amn)
  { int i, j;
    for (i = 0; i < m; i++)
      { for (j = 0; j < n; j++) 
          { Amn[n*i + j] = 2.0 * drandom() - 1.0; }
      }
  }

void throw_LT_matrix(int m, double *Lmm)
  { int i, j;
    for (i = 0; i < m; i++)
      { for (j = 0; j < m; j++) 
          { Lmm[m*i + j] = (j <= i ? 2.0 * drandom() - 1.0 : 0.0); }
      }
  }

void check_simplex(int n, double V[])
  { if (n < 1) { return; }
    double tol = 1.0e-6;
    double dMax = -MAXDOUBLE;
    double dMin = +MAXDOUBLE;
    double rMax = -MAXDOUBLE;
    double rMin = +MAXDOUBLE;
    int i, j, k; 
    for (i = 0; i <= n; i++)
      { /* Check distance from corner {i} to origin: */
        double r2 = 0;
        for (k = 0; k < n; k++)
          { double rk = V[i*n + k]; r2 += rk*rk; }
        double r = sqrt(r2);
        if (r < rMin) { r = rMin; }
        if (r > rMax) { r = rMax; }
        for (j = 0; j < i; j++)
          { /* Check distance from corner {i} to corner {j}: */
            double d2 = 0;
            for (k = 0; k < n; k++)
              { double dk = V[i*n + k] - V[j*n + k]; d2 += dk*dk; }
            double d = sqrt(d2);
            if (d < dMin) { d = dMin; }
            if (d > dMax) { d = dMax; }
          }
      }
    if (rMax - rMin > tol*rMax)
      { fprintf(stderr, "    rMin =       %22.16e\n", rMin);
        fprintf(stderr, "    rMax =       %22.16e\n", rMax);
        fprintf(stderr, "    tolerance =  %22.16e\n", tol);
        demand(FALSE, "** test simplex has irregular radius");
      }
    if (dMax - dMin > tol*dMax)
      { fprintf(stderr, "    dMin =       %22.16e\n", dMin);
        fprintf(stderr, "    dMax =       %22.16e\n", dMax);
        fprintf(stderr, "    tolerance =  %22.16e\n", tol);
        demand(FALSE, "** test simplex has irregular sides");
      }
  }

