/* See enum_orbits.h. */
/* Last edited on 2007-02-06 20:58:31 by stolfi */

#define enum_orbits_C_copyright \
  "Copyright  1996, 2006 Institute of Computing, Unicamp."

/* AUTHORS

  These enumeration procedures were originally developed as the
  Modula-3 file {Oct.m3} by J. Stolfi and Rober M. Rosi in 1993. The
  latter was converted to C by J. Stolfi in 1996 and was substantially
  revised by him in January 2007. */

#include <enum_orbits.h>

#include <vec.h>
#include <bool.h>
#include <ref.h>
#include <ref_set.h>

#include <malloc.h>

bool_t enum_cycle(ref_t p, enum_step_t step, enum_visit_t visit, ref_vec_t *vP)
  { int nP = 0; /* Visited places are {vP->el[0..nP-1]}, if {vP != NULL}. */ 
    ref_t q = p;
    bool_t stop = FALSE;
    do 
      { stop = ( visit == NULL ? FALSE : visit(q));
        if (vP != NULL) { ref_vec_expand(vP, nP); vP->el[nP] = q; nP++; }
        q = step(q);
      } 
    while ((q != p) && (! stop));
    if (vP != NULL) { ref_vec_trim(vP, nP); }
    return stop;
  }

#define INIT_QUEUE_SIZE 1024

bool_t enum_orbits
  ( ref_vec_t root, 
    uint ns, 
    enum_step_t step[], 
    enum_visit_t visit[], 
    ref_vec_t *vP
  )
  {
    ref_vec_t Q = (vP != NULL ? *vP : ref_vec_new(INIT_QUEUE_SIZE));
    int nQ = (vP != NULL ? vP->nel : 0);
      /* The places visited so far (including pre-existing entries
        in {vP}, f any) are {Q[0..nQ-1]}, in stratified order. */

    auto void init_marking(void);
      /* Initializes the tables used by {Marked} and {Mark}. */
    
    auto bool_t check_and_set_mark(ref_t p);
      /* If {p} is not marked as visited, marks {p} and returns FALSE.
        If {p} is already marked, returns TRUE. */
    
    auto void done_marking(void);
      /* Clears any permanent marks and reclaims storage used by
        {is_marked} and {mark}. */
    
    bool_t stop = FALSE; /* Set to TRUE if any {visit[k]} returns TRUE. */

    auto void enum_one(ref_t p, int n);
      /* If {p} is already marked, do nothing. If {p} is unmarked, the
        procedure assumes that {p} was reached through {step[n]} (or,
        if {n==nst}, that {p} is a root). The procedure then marks and
        appends to {Q} the node {p} and every node {q} that can be
        reached from {p} by chains of one or more steps in
        {step[0..n-1]}, visiting each {q} with {visit[k]}, where {k}
        is the indes of the last step used to reach {q}. Stops as soon
        as a {visit[k]} procedure returns TRUE. */
    
    auto void close_orbit(int n);
      /* Finds all unmarked items that can be reached from item
        {Q[nQ-1]} by chains of steps in {step[0..n-1]}; marks every
        such item {q}, appends {q} to {Q[0..nQ-1]}, and visits {q}
        with {visit[k]} where {k} is the index of the last step used
        to reach {q}. Stops as soon as a {visit[k]} procedure returns
        TRUE. */
    
    /* Initialize the mark tables: */
    init_marking();
    
    /* Enumerate all roots: */
    int i;
    for (i = 0; (i < root.nel) && (! stop); i++)
      { enum_one(root.el[i], ns); }

    /* Release the marking table storage: */
    done_marking();
    
    /* Return {Q}, or reclaim its space: */
    if (vP == NULL)
      { free(Q.el); }
    else
      { ref_vec_trim(&Q, nQ); (*vP) = Q; }
      
    /* Return the aborted-enum indicator: */
    return stop;

    /* MAIN PROCEDURE */
    
    void enum_one(ref_t p, int n)
      { if (! check_and_set_mark(p))
          { /* New item, stack it: */
            ref_vec_expand(&Q, nQ); 
            Q.el[nQ] = p; nQ++;
            /* Visit it: */
            if (visit[n] != NULL) 
              { stop = visit[n](p); if (stop) { return; } } 
            /* Gather its orbit under {step[0..n-1]}: */
            close_orbit(n);
          }
      }
    
    void close_orbit(int n)
      { if (n > 0)
          { int nX = nQ - 1;
            close_orbit(n-1);
            while((nX < nQ) && (! stop))
              { ref_t q = Q.el[nX]; nX++;
                q = step[n-1](q);
                enum_one(q, n-1);
              }
          }
      }
             
    /* MARKING PROCEDURES */
    
    ref_set_t M;
    uint nM; 
      /* The set {M} contains the addresses visited so far.
        The variable {nM} keeps track of the number of elements in {M}. */
    
    void init_marking(void)
      { M = ref_set_new(INIT_QUEUE_SIZE);
        nM = 0; /* No wedges have been seen yet. */
      }
    
    bool_t check_and_set_mark(ref_t p)
      { bool_t was_there = ref_set_add(&M, p);
        if (! was_there) 
          { nM++;
            if (2*nM > M.nel) { ref_set_resize(&M, 2*M.nel+1); }
          }
        return was_there;
      }
    
    void done_marking(void)
      { free(M.el); }
  }
