/* See ref_set.h */
/* Last edited on 2007-02-05 18:54:59 by stolfi */

#include <ref_set.h>

#include <vec.h>
#include <bool.h>
#include <ref.h>
#include <jsrandom.h>
#include <affirm.h>

#include <stdlib.h>
#include <assert.h>

unsigned int ref_set_locate(ref_set_t *S, ref_t a);
  /* If {a} is in the set {S}, returns the index {k} such that
    {S->el[k] == a}.
    
    If {a} is not in {S}, and {S} is not full, returns the index {k},
    of the slot {S->el[k]} where {a} should be stored to
    add it to the set; in that case, {S[k] == NULL}.
    
    If {a} is not in {S} but {S} is full, returns some 
    index {k} in {0..S->nel-1}, such that {S->el[k]}
    is neither {a} nor {NULL}.
    
    Assumes that {S->nel > 0} and {a != NULL}. */

unsigned int ref_set_locate(ref_set_t *S, ref_t a)
  { unsigned int n = S->nel;
    unsigned int k = ref_set_hash(a, n);
    unsigned int i = k;
    do { 
      ref_t b = S->el[i];
      if ((b == a) || (b == NULL)) { return i; }
      i = (i + 1) % n;
    } while (i != k);
    return k;
  }

ref_set_t ref_set_new(int n)
  { ref_set_t S = ref_vec_new(n);
    ref_set_clear(&S);
    return S;
  }

void ref_set_clear(ref_set_t *S)
  { unsigned int i;
    for (i = 0; i < S->nel; i++) { S->el[i] = NULL; }
  }

void ref_set_resize(ref_set_t *S, int n)
  { if (S->nel != n) 
      { /* Allocate a new address vector with {n} slots: */
        ref_set_t R = ref_set_new(n);
        /* Insert all elements of {S} into {R}: */
        unsigned int i;
        for (i = 0; i < S->nel; i++)
          { ref_t a = S->el[i];
            if (a != NULL) { ref_set_add(&R, a); }
          }
        /* Reclaim the old vector and replace it by the new one: */
        free(S->el); (*S) = R;
      }
  }

bool_t ref_set_has(ref_set_t *S, ref_t a)
  { /* Argument checking: */
    demand(a != NULL, "invalid element");
    /* Get the index {k} of {a} in {S}: */
    unsigned int k = ref_set_locate(S, a);
    return (S->el[k] == a);
  }

bool_t ref_set_add(ref_set_t *S, ref_t a)
  { /* Argument checking: */
    demand(a != NULL, "invalid element");
    /* Trivial case: */
    demand(S->nel > 0, "no space for any elements");
    /* Get the index {k} of {a} in {S}: */
    unsigned int k = ref_set_locate(S, a);
    /* Special cases: */
    if (S->el[k] == a) { /* {a} is already in {S} */ return TRUE; }
    demand(S->el[k] == NULL, "no space for new element");
    S->el[k] = a;
    return FALSE;
  }

bool_t ref_set_del(ref_set_t *S, ref_t a)
  { /* Argument checking: */
    demand(a != NULL, "invalid element");
    /* Trivial case: */
    if (S->nel == 0) { return FALSE; }
    /* Get the index {k} of {a} in {S}: */
    unsigned int k = ref_set_locate(S, a);
    if (S->el[k] != a) { /* {a} is not in {S} */ return FALSE; }
    /* Erase the entry: */
    S->el[k] = NULL;
    /* Now entries that follow {S[k]} need to be deleted and reinserted: */
    unsigned int i = k;
    while(TRUE)
      { i = (i + 1) % S->nel;
        ref_t a = S->el[i];
        if (a == NULL) { return TRUE; }
        S->el[i] = NULL;
        bool_t found = ref_set_add(S, a);
        assert(! found);
      }
  }

ref_t ref_set_get(ref_set_t *S)
  { if (S->nel == 0) { return NULL; }
    unsigned int k = abrandom(0, S->nel-1);
    unsigned int i = k;
    do {
      ref_t a = S->el[i];
      if (a != NULL) { return a; }
      i++;
    } while (i != k);
    return NULL;
  }

unsigned int ref_set_count(ref_set_t *S)
  { unsigned int n = 0;
    unsigned int i;
    for (i = 0; i < S->nel; i++) 
      { if (S->el[i] != NULL) { n++; } }
    return n;
  }

ref_vec_t ref_set_elems(ref_set_t *S)
  { /* Get the element count {n = #S}: */
    unsigned int n = ref_set_count(S);
    /* Allocate the result vector: */
    ref_vec_t v = ref_vec_new(n);
    /* Scan {S} and copy the nin-{NULL} entries to {v}: */
    n = 0;
    unsigned int i;
    for (i = 0; i < S->nel; i++)
      { ref_t a = S->el[i];
        if (a != NULL) { v.el[n] = a; n++; }
      }
    return v;
  }
  
unsigned int ref_set_hash(ref_t a, unsigned int n)
  { demand(a != NULL, "cannot hash NULL");
    demand(n > 0, "cannot hash to empty range");
    /* Let's hope that this works: */
    unsigned int ua = (unsigned int) a;
    unsigned int ub = ua & 8191u;
    unsigned int ur = (4615u * (ua >> 14)) + (417 * (ua >> 9)) + (ub * ub * 471703u);
    return (ur % n);
  }
