#ifndef ref_set_H
#define ref_set_H

/* Sets of addresses, using sequential hashing. */
/* Last edited on 2007-02-05 18:54:39 by stolfi */

#include <vec.h>
#include <ref.h>
#include <bool.h>

typedef ref_vec_t ref_set_t;
  /* A {ref_set_t} is a set of non-NULL addresses ({ref_t} values).
  
    The set is represented by a vector {S} of {void*} pointers,
    {S.el[0..S.nel-1]}. The elements of the set are the non-NULL
    entries of {S}. Each new element {a} added to the set is stored in
    the first NULL slot found starting at position {ref_set_hash(a)},
    wrapping around modulo {S.nel}.
    
    The efficiency of the set operations below depends on the
    occupancy ratio {r == #S/S.nel}, where {#S} is the number
    of (non-NULL) elements in the set. 
    
    The functions {ref_set_has}, {ref_set_add}, and {ref_set_del} are
    efficient as long as {r} is well below 1. In that case, their
    expected cost is {O(1)} operations, independently of {#S} or
    {S.nel}. However, the cost grows very quickly, to O(S.nel) or
    more, as {r} approaches 1.
    
    On the other hand, the expected cost of {ref_set_get} is {O(1)} if
    {r} is well above 0, but grows quickly to O(S.nel) as {r} goes to zero.
    
    The client should therefore keep track of {#S}, and call
    {ref_set_resize} as needed to keep the occupancy ratio {r} within
    some range {[a _ b]} with {0 << a < b << 1}. The best values of
    {a} and {b} depend on the situation, but {a==1/4} and {b==1/2}
    should be good enough for most applications. */

ref_set_t ref_set_new(int n); 
  /* Creates a new {ref_set_t} with {n} slots, initially empty.
    Cost: {O(n)}. */

void ref_set_clear(ref_set_t *S); 
  /* Makes {S} empty, i.e. deletes all elements from {S}.
    Cost: {O(S->nel)}. */

void ref_set_resize(ref_set_t *S, int n);
  /* Resizes {S} to have exactly {n} entries, and re-inserts
    all existing elements.  Fails if {S} already contains
    more than {n} elements.  Cost: {O(n + S->nel)}. */
    
bool_t ref_set_has(ref_set_t *S, ref_t a);
  /* Returns TRUE if {a} is in {S}, FALSE if it isn't. \
    Fails if {a} is NULL. Cost: expected {O(1)} if {S} is not too
    full; {O(S->nel)} otherwise. */
    
bool_t ref_set_add(ref_set_t *S, ref_t a);
  /* If {a} is not in the set {S}, adds it and returns FALSE.
    If {a} is already in {S}, does nothing and returns TRUE.
    Fails if {a} is NULL, or if {a} is not in {S} and
    {S} already contains {S.nel} non-NULL elements. 
    Cost: expected {O(1)} if {S} is not too full; {O(S->nel)}
    otherwise. */
    
bool_t ref_set_del(ref_set_t *S, ref_t a);
  /* If {a} is in the set {S}, deletes it and returns TRUE.
    If {a} is not in {S}, does nothing and returns FALSE.
    Fails if {a} is NULL. Cost: expected {O(1)} if not too
    full; {O((S->nel)^2)} otherwise. */    

ref_t ref_set_get(ref_set_t *S);
  /* Returns a random element of {S}, or NULL if {S} is empty.
    Cost: {O(1)} is {S} is not to empty; {O(S.nel)} otherwise. */    

unsigned int ref_set_count(ref_set_t *S);
  /* The number {#S} of elements in {S}, i.e. the number of non-NULL
    entries in {S.el[0..S.nel-1]}. 
    Cost: always {O(S.nel)}, independently of {#S}. */

ref_vec_t ref_set_elems(ref_set_t *S);
  /* Returns a vector that contains all the non-NULL entries
    in {S}, with no {NULL}s, in arbitrary order. 
    Cost: always {O(S.nel)}, independently of {#S}. */

unsigned int ref_set_hash(ref_t a, unsigned int n);
  /* Returns a hash value in {0..n-1}, computed from {a}.
    Fails if {a} is NULL or {n} is 0.
    Cost: {O(1)}. */

#endif
