#define PROG_NAME "testrefset"
#define PROG_DESC "tests the {ref_set.h} procedures"
#define PROG_VERS "1.1"

/* Last edited on 2007-02-05 19:09:02 by stolfi */
/* Created on 2007-01-31 by J. Stolfi, UNICAMP */

#define PROG_COPYRIGHT \
  "Copyright  2007  by the State University of Campinas (UNICAMP)"

#include <ref_set.h>

#include <vec.h>
#include <bool.h>
#include <ref.h>
#include <affirm.h>
#include <jsrandom.h>
#include <jsmath.h>
#include <jstime.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <values.h>

int main (int argc, char **argv);

void test_ref_set_static(int nElems, int nSlots, int nTimes, int szMin, int szMax);
  /* Tests the {ref_set.h} operations on a set vector with static size
    {nSlots} and containing at most {nElems} elements. The timing
    loops execute each operation about {nTimes} times. The test
    addresses are obtained by allocating {nElems} heap records with
    sizes in the range {szMin..szMax}. */

void test_ref_set_correctness(int nElems, ref_t elem[], ref_set_t *S);
void test_ref_set_speed(int nElems, ref_t elem[], ref_set_t *S, int nTimes);
void print_timing(char *func, double usec, int nops);

int main (int argc, char **argv)
  {
    srandom(4615);
    
    int nTimes = 1000000; /* Number of timing calls per function. */
    double maxOcc = 0.50; /* Max occupancy of set. */
    int nElems;
    for (nElems = 100; nElems <= 10000; nElems *= 10)
      { int nSlots = (int)(ceil(nElems/maxOcc)); /* Number of slots in set vector. */
        test_ref_set_static(nElems, nSlots, nTimes,    1,    1);
        test_ref_set_static(nElems, nSlots, nTimes,  100,  200);
        test_ref_set_static(nElems, nSlots, nTimes,    1, 1000);
      }
    return 0;
  }
  
#define MAX_ELEM_BYTES 10000000
   /* Maximum memory to allocate for tests records. */
  
void test_ref_set_static(int nElems, int nSlots, int nTimes, int szMin, int szMax)
  { 
    fprintf(stderr, "============================================================\n");
    fprintf(stderr, "testing with %d elements in %d slots", nElems, nSlots);
    fprintf(stderr, " (occupancy ratio %6.4f)\n", ((double)nElems)/((double)nSlots));
    fprintf(stderr, "record size range = [%d .. %d]\n", szMin, szMax);
    
    int i;
    
    /* Sanity check for memory size: */
    double expmem = ((double)(szMin + szMax))/2*nElems;
    demand(expmem < MAX_ELEM_BYTES, "test requires for too much memory");
    
    /* Create the addresses to store in the set: */
    fprintf(stderr, "creating the elements ...\n");
    ref_t elem[nElems];
    for (i = 0; i < nElems; i++)
      { size_t sz = abrandom(szMin, szMax);
        elem[i] = notnull(malloc(sz), "no mem");
      }
      
    /* Allocate the set for the specified occupancy ratio: */
    fprintf(stderr, "testing {ref_set_new} ...\n");
    ref_set_t S = ref_set_new(nSlots);
    affirm(S.nel == nSlots, "wrong vector size");
    affirm(S.el != NULL, "no storage area");
    
    test_ref_set_correctness(nElems, elem, &S);
    
    test_ref_set_speed(nElems, elem, &S, nTimes);
    
    /* Reclaim test storage: */
    free(S.el);
    for (i = 0; i < nElems; i++) { free(elem[i]); }
    
    fprintf(stderr, "============================================================\n");
    return;
  }
  
void test_ref_set_correctness(int nElems, ref_t elem[], ref_set_t *S)
  {
    fprintf(stderr, "TESTING CORRECTNESS\n");
    
    bool_t found;
    int i;
    
    /* Add the elements to the set: */
    fprintf(stderr, "testing {ref_set_add}, {ref_set_has} ...\n");
    for (i = 0; i < nElems; i++)
      { ref_t a = elem[i];
        affirm(! ref_set_has(S, a), "{ref_set_has} error 1 (found when shouldn't)");
        found = ref_set_add(S, a);
        affirm(! found, "{ref_set_add} error 1 (claims that it was already there)");
        affirm(ref_set_has(S, a), "{ref_set_has} can't find element added with {ref_set_add}");
        found = ref_set_add(S, a);
        affirm(found, "{ref_set_add} error 2 (claims that it was not there)");
      }
    
    /* Try deleting and reinserting each element: */
    fprintf(stderr, "testing {ref_set_del} ...\n");
    for (i = 0; i < nElems; i++)
      { ref_t a = elem[i];
        affirm(ref_set_has(S, a), "{ref_set_has} error 3 (not found when should)");
        found = ref_set_del(S, a);
        affirm(found, "{ref_set_del} error 1 (claims that element is not there)");
        affirm(! ref_set_has(S, a), "{ref_set_has} found a deleted element");
        found = ref_set_add(S, a);
        affirm(! found, "{ref_set_add} error 1 (claims that it was already there)");
        affirm(ref_set_has(S, a), "{ref_set_has} error 4 (not found when should)");
      }
      
    /* Make sure that all elements are still in {S}: */
    fprintf(stderr, "checking again ...\n");
    for (i = 0; i < nElems; i++)
      { ref_t a = elem[i];
        affirm(ref_set_has(S, a), "{ref_set_has} error 5 (not found when should)");
      }
  }

void test_ref_set_speed(int nElems, ref_t elem[], ref_set_t *S, int nTimes)
  {
    fprintf(stderr, "TESTING SPEED\n");
    
    bool_t found;
    int kc, i;
    
    /* Pick a number {step} that is relatively prime to {nElems}: */
    int step = (nElems + 1)/2; 
    while(gcd(step, nElems) != 1) { step--; }

    /* Do some timing. We remove 1/2 of the elements, then add them back: */
    int nBatch = nElems/2; /* Number of elements to remove/add per cycle. */
    assert(nBatch <= nElems);
    int nCycles = (nTimes + nBatch - 1)/nBatch; /* Number of cycles. */
    assert(nCycles > 0);
    
    int kDel = 0; /* Next element to delete. */
    int kAdd = 0; /* Next element to add back. */
    int kHas = nElems/2; /* Next element to check for presence. */
    double tAdd = 0, tDel = 0, tHas = 0, tGet = 0;
    int nAdd = 0, nDel = 0, nHas = 0, nGet = 0;
    double start, stop;
    for (kc = 0; kc < nCycles; kc++)
      { 
        /* Delete {nBatch} elements: */
        start = user_cpu_time_usec();
        for (i = 0; i < nBatch; i++)
          { ref_t a = elem[kDel]; kDel = (kDel + step) % nElems; 
            found = ref_set_del(S, a); assert(found);
          }
        stop = user_cpu_time_usec();
        tDel += stop - start; nDel += nBatch; 
        
        /* Add them back: */
        start = user_cpu_time_usec();
        for (i = 0; i < nBatch; i++)
          { ref_t a = elem[kAdd]; kAdd = (kAdd + step) % nElems; 
            found = ref_set_add(S, a); assert(! found);
          }
        stop = user_cpu_time_usec();
        tAdd += stop - start; nAdd += nBatch; 
        
        /* Check their presence: */
        start = user_cpu_time_usec();
        for (i = 0; i < nBatch; i++)
          { ref_t a = elem[kHas]; kHas = (kHas + step) % nElems; 
            found = ref_set_has(S, a); assert(found);
          }
        stop = user_cpu_time_usec();
        tHas += stop - start; nHas += nBatch; 
        
        /* Get some elements: */
        start = user_cpu_time_usec();
        for (i = 0; i < nBatch; i++)
          { ref_t a = ref_set_get(S); assert(a != NULL); }
        stop = user_cpu_time_usec();
        tGet += stop - start; nGet += nBatch; 
      }
          
    fprintf(stderr, "\n");
    fprintf(stderr, "times\n");
    print_timing("ref_set_add", tAdd, nAdd);
    print_timing("ref_set_del", tDel, nDel);
    print_timing("ref_set_has", tHas, nHas);
    print_timing("ref_set_get", tGet, nGet);
  }

void print_timing(char *func, double usec, int nops)
  {
    fprintf(stderr, "%-20s  %13.0f usec / %10d ops = %13.6f usec/op\n", func, usec, nops, usec/nops);
  }
