/* pswr_aux.h -- internal prototypes for pswr.c. */
/* Last edited on 2007-01-04 02:50:23 by stolfi */

#ifndef pswr_aux_H
#define pswr_aux_H

#include <pswr.h>
#include <bool.h>

#include <stdio.h>

/* STREAM STATUS */

/* A stream that has been created but not been closed can be in two
  states: 
  
    * between_pages ({curPage >= 0},{curSlot = INT_MAX}), or 
    
    * within_page ({curPage > 0}, {curSlot \in -1..hCount*vCount-1}). 
  
  For a standalone document, these states mean:

    * between_pages: the file {psFile} is open and the file preamble
      has been written out to it. The page preamble has not been
      written yet.

    * within_page: the file is open and both the file and page
      preambles have been written.

  For an encapsulated stream, the states mean:

    * between_pages: the file {psFile} is NULL.

    * within_page: the encapsulated picture file is open and all
      necessary preambles have been written to it.
      
  A newly-created stream is in the between_pages state. The
  transitions between these states are effected by the procedures
  {pswr_begin_page} and {pswr_end_page}.  Drawing commands may be
  issued only in the {within_page} state. */

bool_t pswr_within_page(PSStream *ps);
  /* TRUE between a {pswr_begin_page} and the next {pswr_end_page}. */

bool_t pswr_next_slot_available(PSStream *ps);
  /* TRUE if there is some unused picture slot in the current page,
    besides the current one. */

bool_t pswr_page_dirty(PSStream *ps);
  /* TRUE if {pswr_new_picture} or {pswr_set_window} were issued
    to the current page. */
    
/* PAGE OPEN AND CLOSE */

void pswr_begin_page(PSStream *ps, const char *pageName);
  /* Starts a new logical page of {ps}, and sets the current
    picture slot to -1. Must be followed eventually by a call to
    {pswr_end_page}. */

void pswr_end_page(PSStream *ps);
  /* Terminates the current logical page. Must be followed by
    {pswr_begin_page} or {pswr_close_stream}. */

/* PAGE LAYOUT */

void pswr_fix_slot_size
  ( double usableSize, 
    double picMargin,
    double capSize,
    int *count,
    double *picSize, 
    double *slotSize
  );
  /* Adjusts {*count}, {*picSize}, and {*slotSize}, if needed,
    given the parameters {usableSize}, {picMargin}, and {capSize}.
    
    The paramters are assumed to refer to a string of picture slots
    arranged in either horizontal or vertical direction. We will
    assume vertical, since horizontal is analogous except that
    {capSize} is probably zero. All parameters are taken to be in pt.
    
    The {usableSize} argument is the total heigh of the page minus the
    page margins. The {picMargin} is the amount of blank space that
    must be allowed on either side of each picture slot, except before
    the first slot and after the last one. The {capSize} is the extra
    space, for caption text, that must be allowed below each slot. The
    number of pictures in the string is {*count}. The size of a
    picture area, excluding all margins, is {*picSize}; and the
    distance between successive slots is {*slotSize}.
  
    If {count <= 0} and {picSize <= 0}, then count is set to 1 
    and {picSize} is set to {usableWidth}. */

/* ENCAPSULATED FIGURES */

void pswr_begin_encapsulated_figure(PSStreamPrivate *ps);
  /* Initializes the underlying stream as an Encapsulated PostScript file.
    
    Writes the file's preamble, which defines some auxiliary
    Postscript operators and constants, sets up the caption font,
    resets some graphics variables (line width and color, fill color,
    etc.), etc. The figure's bounding box (for the "%%BoundingBox"
    comment) is taken from the fields {ps->hPageSize,ps->vPageSize}.
    The plotting window is initialized to cover the whole figure, with
    client coordinates equal to the device coordinates. */
    
void pswr_end_encapsulated_figure(PSStreamPrivate *ps);
  /* Finalizes an Encapsulated PostScript figure. The file
    is flushed but is left open. */

/* NON-ENCAPSULATED MULTIPAGE DOCUMENTS */

void pswr_begin_document(PSStreamPrivate *ps);
  /* Initializes a file to contain a non-encapsulated, possibly
    multi-page Postscript document.
    
    The procedure writes the file preamble, which defines some
    auxiliary Postscript operators and constants, defines the 
    caption font, etc.
    
    The client must initialize each page with {ps_begin_document_page}, before
    issuing any drawing commands, and finalize it with {ps_end_document_page}. */

void pswr_begin_document_page(PSStreamPrivate *ps, const char *page);
  /* Starts a new page in a non-encapsulated document.  
  
    The {page} string is the logical page number; if not NULL and
    non-empty, it will be printed at the bottom of the page. It is is
    also used as the first argument of the "%%Page" structuring
    directive (where it defaults to the sequential page number,
    counting from 1). 
    
    At each new page, the plotting window is reset to the whole page, 
    and the whole Postscript state (including line width and color, fill color,
    etc.) is restored to some default state. */

void pswr_end_document_page(PSStreamPrivate *ps);
  /* Finalizes a page in a non-encapsulated document. */

void pswr_end_document(PSStreamPrivate *ps);
  /* Finalizes a multipage document.  The file is flushed but left open. */
    
/* INTERNAL SCALING/CLIPPING PROCEDURES */

void pswr_save_default_window_data
  ( PSStream *ps,
    double hmin, double hmax,
    double vmin, double vmax
  );
  /* Saves the given data as the current plot rectangle
    ({ps->hmin,ps->hmax,ps->vmin,ps->vmax}), and also as the current
    client coordinate ranges ({ps->xmin,ps->xmax,ps->ymin,ps->ymax}). Sets
    the current cell grid ({ps->xn,ps->yn}) to a single cell. */
    
void pswr_save_window_data
  ( PSStream *ps,
    double xmin, double xmax,
    double ymin, double ymax,
    double hmin, double hmax,
    double vmin, double vmax,
    int xn, int yn
  );
  /* Saves the given data as the current plot rectangle
    ({ps->hmin,ps->hmax,ps->vmin,ps->vmax}), current client coordinate
    ranges ({ps->xmin,ps->xmax,ps->ymin,ps->ymax}), and current cell grid
    ({ps->xn,ps->yn}). */
    
/* FILE PREMBLES AND POSTAMBLES */

/* The file preamble (produced by both pswr_write_eps_file_header and
  pswr_write_ps_file_header) creates a Postscript dictionary
  "ps$dict" that contains definitions for special Postscript
  operators and some state variables, such as the plotting rectangle
  and grid size. This dictionary is closed before the end of the
  preamble, and must be re-opened at each page. */

void pswr_write_eps_file_header
  ( FILE *psfile, 
    const char *date,
    double hmin, double hmax,
    double vmin, double vmax
  );
  /* Writes the file preamble for an encapsulated Postscript file. The
    figure's "%%BoundingBox" will be set to the given values (in pt),
    rounded outwards to integers. Must be followed by
    pswr_write_page_header to complete the setup. */

void pswr_write_eps_file_trailer(FILE *psfile, int nfonts, char **fonts);
  /* Writes the file postamble for an encapsulated Postscript file,
    namely the "%%Trailer" comment followed by the "%%DocumentFonts:
    {F}" (where {F} is the given list of font names). This call must
    be preceded by pswr_write_page_trailer. */

void pswr_write_ps_file_header(FILE *psfile, const char *papersize, const char *date);
  /* Writes the file preamble (including operator definitions) for a
    non-encapsulated Postscript document file, that specifies the
    given paper size, ending with "%%EndProlog". Should be followed by
    pswr_write_page_header. */
    
void pswr_write_ps_file_trailer(FILE *psfile, int npages, int nfonts, char **fonts);
  /* Writes the trailer for a non-encapsulated Postscript document.
    The trailer consists of a "%%Trailer" line, plus the directives
    "%%Pages: {N}" (where {N} is the given page count) and
    "%%DocumentFonts: {F}" (where {F} is the given list of font
    names). This call must be preceded by pswr_write_page_trailer. */
    
/* DOCUMENT FONT LIST */

void pswr_initialize_font_list(int *nfontsP, char ***fontsP);
  /* Initializes the font list with the "Courier" font. */

void pswr_add_font(const char *font, int *nfontsP, char ***fontsP);
  /* Appends a copy of the given {font} string to the font list. */

void pswr_free_font_list(int *nfontsP, char ***fontsP);
  /* Frees all storage used by the font list and sets it to NULL. */

/* PAGE PREAMBLES AND POSTAMBLES */

void pswr_write_page_header(FILE *psfile, const char *pagename, int pageseq);
  /* Writes the preamble for a new page of a Postscript file (EPS or
    non-EPS; if the former, may be called only once). The page
    preamble includes the "%%Page" directive, and commands that save
    the current Postscript state (with "save"), and open the
    "ps$dict" dictionary. */
    
void pswr_write_page_trailer(FILE *psfile, int eps);
  /* Writes the postamble for a page in a Postscript file (EPS or
    non-EPS). 
    
    The page postamble includes a "showpage" command, if {eps} is
    false; then commands that close the "ps$dict" dictionary, and
    restore the Postcript state that was saved when the page was
    started. */

/* MISCELLANEOUS WRITING PROCEDURES */

void pswr_write_proc_defs(FILE *psfile);
  /* Writes to {psfile} the definition of the operators assumed
    by other plotting and text writing commands. */
  
void pswr_write_date_and_page_show_cmds
  ( FILE *psfile, 
    const char *date, 
    const char *page
  );
  /* Writes to {ps->psfile} some commands that print the 
    given date and page number near the bottom of the
    current page. */

void pswr_write_page_graphics_setup_cmds(FILE *psfile, double *fc);
  /* Writes to {ps->psfile} some commands that setup some graphics
    variables, suitable for the beginning of a new page or figure.
    In particular the stroke color is reset to black, and the 
    fill color to {fc}. */
  
void pswr_write_window_setup_cmds
  ( FILE *psfile,
    double hmin, double hmax,
    double vmin, double vmax,
    int xn, int yn
  );
  /* Writes to {psfile} some commands that setup the Postscript
    variables {xmin,xmax,ymin,ymax} to the values
    {hmin,hmax,vmin,vmax}, and set the clip path to that rectangle.
    Those commands also set the Postscript variables {xn,yn} that
    define the cell grid; {xtext,ytext}, which define
    the start of the current caption line; and 
    {dytext} that defines the caption interline
    spacing (in pt).  Finally, they set the {captionfont}
    varriable to the caption font, properly scaled. */

void pswr_write_label_font_setup_cmds(FILE *psfile, const char *font, double size);
  /* Writes to {psfile} commands that create an instance of the font
    named {font}, scaled by {size} pt, and save it in the Postscript
    variable {labelfont} (which is used by ps_label). */

void pswr_write_ps_string(FILE *psfile, const char *text, const char *newline);
  /* Writes the given text out  to /psfile/, in Postscript form, properly
    quoting any special chars and parentheses. 
    
    Replaces any embedded '\n' by the given /newline/ string --- which
    could be, for example, ") show mynl (". */

void pswr_write_color(FILE *psfile, double *clr);
  /* Writes the color value {clr[0..2]} to the Postscript file,
    with 3 decimals, preceded by spaces.  */

void pswr_write_font_list(FILE *psfile, int nfonts, char **fonts);
  /* Writes the structured comment "%%DocumentFonts: {F}" 
    where {F} is the given list of font names. */

void pswr_write_fill_color_set_cmd(FILE *psfile, double *fc);
  /* Writes the Postoscript command that saves {fc[0..2]}
    as the RGB fill color for subsequent figure commands. */

#endif
